/* clock.cc       last Update: 23.8.2012
accurate clock with Timer1
 Version: 0.1
 Prozessor: Atmega8
 Circuit: Port B0,B1,B2 connected to LEDs (over 470 Ohm resistors)
 Autor: Rolf Pfister
 Copyright: Freeware

 History:
 5.5.2009	Erstellung, first version in german
 23.8.2012      translated to english
                names of interruptroutine updated to new convention
		some variables updated to more meaningful names

*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "ulong.h" //definition of uchar, uint, and ulong

/**** definitions for timer ****/
static volatile uint millisec=0,days=0;
static volatile uchar sec=0,min=0,hours=0;

// measured quartz frequency in Hertz and thousandth of Hertz:

#define F_CPU 3686400UL //exactly 3.6864 MHz (UL means unsigned long)
#define MilliHz 0

//Example of measured frequency: with exactly 3.6864 MHz after 10 hours the clock was 0.35 sec late:
// so the frequency was a bit lower, calculated in ticks per second:
// 3686400 * 0.35 / (3600*10) = 35.84
// corrcted frequency: 3686400-35.84 = 3686364.160
//#define F_CPU 3686364UL //measured frequency
//#define MilliHz 160     //measured frequency

const uint Rest1 = F_CPU%1000; //every second needed correction
const uint Rest2 = MilliHz*6/100; //every minute needed correction
const uint Rest3 = ((MilliHz*6-Rest2*100)*60+50)/100; //every hour needed correction
const uint NormalerCompwert=F_CPU/1000-1;

//the names SIG_* can be found in include/avr/iom8.h
//SIGNAL (SIG_OUTPUT_COMPARE1A) //old style of interrupt-routines, dont use anymore

ISR(TIMER1_COMPA_vect)
{
 uchar mehr;
 uint ms=millisec;
 if(++ms>=1000-Rest1) mehr=1; else mehr=0;
 if(ms==1000)
  {ms=0;
   if(++sec>=60-Rest2) mehr=1; else mehr=0;
   if(sec==60)
    {sec=0;
     if(++min>=60-Rest3) mehr=1; else mehr=0;
     if(min==60)
      {min=0; mehr=0;
       if(++hours==24)
	{hours=0; ++days;}
      }
    }
  }
 millisec=ms;
 OCR1A=NormalerCompwert+mehr;
}

void milliwait(int n) //wait n milliseconds, n must be between 1 and 999 (including)
{
 uint ms;
 //ms=millisec+n;  this would the compiler not make really correctly!
 cli(); ms=millisec; sei(); ms+=n;
 if(ms>=1000) ms-=1000;
 while(ms!=millisec || ms!=millisec) ; //wait until new time really reached
}

int main(void)
{
 int n=0;
 DDRB  = 0xFF; //PortB PB0-3 defining as output
 PORTB = n; //all LEDs off

 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1
 OCR1A = NormalerCompwert; //compare value
 TCNT1 = 0; //Start value of Timer1
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts at compare value

 sei(); //Interrupts turning on

 //comment out one of the two main loops
 while(1) //main loop
   {
    PORTB=sec; //show running seconds on the LEDs
   }

 /*
 while(1) //main loop
  {
   milliwait(500);
   n+=1;            //count up the LEDs every 0.5 second
   PORTB=n;
  }
 */

 return 0;
}
